require "goliath"

require_relative "./api_transformer/endpoint"
require_relative "./api_transformer/errors"
require_relative "./api_transformer/backend_request"
require_relative "./api_transformer/backend_request_sender"
require_relative "./api_transformer/backend_response"
require_relative "./api_transformer/frontend_response"
require_relative "./api_transformer/frontend_response_builder"
require_relative "./api_transformer/params"
require_relative "./api_transformer/route"
require_relative "./api_transformer/routes"
require_relative "./api_transformer/rack/cookie_params"

module ApiTransformer
  # Inherit from this class to implement an API transformation server:
  #
  #   class ExampleServer < ApiTransformer::Server
  #     base_url "http://ip.jsontest.com/"
  #
  #     get "/ip" do |params|
  #       request :ip do
  #         path "/"
  #         method :get
  #       end
  #
  #       response do |data|
  #         success do
  #           status 200
  #           attribute :your_ip, data[:ip][:ip]
  #         end
  #       end
  #     end
  #   end
  #
  class Server < Goliath::API
    use Goliath::Rack::Params
    use ApiTransformer::Rack::CookieParams

    def on_headers(env, headers)
      env["client-headers"] = headers
    end

    def response(env)
      path = env[Goliath::Request::REQUEST_PATH]
      route, path_params = @@routes.find(env["REQUEST_METHOD"], path)

      if route
        all_params = params.merge(path_params)
        response = route.run(@@base_url, all_params, env)
        streaming_response(response.status, response.headers)
      else
        [404, {}, "nope"]
      end
    end

    class << self
      def inherited(klass)
        klass.use Goliath::Rack::Params
      end

      def base_url(url)
        @@base_url = url
      end

      def namespace(path)
        prior = [@namespace, failure_handlers.dup, helper_blocks.dup]
        @namespace = @namespace.to_s + path

        yield

        @namespace, @failure_handlers, @helper_blocks = prior
      end

      def unhandled_failures(&block)
        failure_handlers.unshift(block)
      end

      def helpers(&block)
        helper_blocks.unshift(block)
      end

      def shared_endpoints(name, &block)
        if shared_endpoint_blocks[name]
          fail "shared_endpoints already defined for \"#{name}\""
        else
          shared_endpoint_blocks[name] = block
        end
      end

      def include_endpoints(name)
        if shared_endpoint_blocks[name]
          instance_eval(&shared_endpoint_blocks[name])
        else
          fail "missing shared_endpoints for \"#{name}\""
        end
      end

      def get(path = "", options = {}, &block)
        add_route(:get, path, options, block)
      end

      def post(path = "", options = {}, &block)
        add_route(:post, path, options, block)
      end

      def put(path = "", options = {}, &block)
        add_route(:put, path, options, block)
      end

      def delete(path = "", options = {}, &block)
        add_route(:delete, path, options, block)
      end

      def reset
        @@routes =  @failure_handlers = @helper_blocks = @shared_endpoints = nil
      end

      private

      def add_route(method, path, options, block)
        full_path = "#{@namespace}#{path}"

        @@routes ||= Routes.new
        @@routes.add(method: method, path: full_path, options: options,
          block: block, failure_handlers: failure_handlers,
          helper_blocks: helper_blocks)
      end

      def failure_handlers
        @failure_handlers ||= []
      end

      def helper_blocks
        @helper_blocks ||= []
      end

      def shared_endpoint_blocks
        @shared_endpoints ||= {}
      end
    end
  end
end
