require "json"
require "hashie"

module ApiTransformer
  # Container for backend response data
  class BackendResponse
    # Subclasses Hash so that it can interchangeably use string or symbol access
    class Data < Hash
      include Hashie::Extensions::IndifferentAccess
    end

    attr_reader :body

    def initialize(http)
      @http = http
      @body = ""

      @http.stream { |chunk| @body += chunk }
    end

    def success?
      status >= 200 && status < 300
    end

    def status
      @http.response_header.http_status
    end

    def json
      @data ||= begin
        json = JSON.parse(@body)

        if json.respond_to?(:keys)
          Data.try_convert(json)
        else
          json
        end
      end
    rescue JSON::ParserError, Encoding::InvalidByteSequenceError
      {}
    end

    def [](key)
      if json.respond_to?(:keys)
        json[key]
      else
        nil
      end
    end

    def cookies
      pairs = @http.response_header["SET_COOKIE"].split("\s*;\s*")
      data = pairs.reduce({}) do |hash, pair|
        key, value = pair.split("=")
        hash.merge(key => value)
      end

      Data.try_convert(data)
    end

    def content_type
      @http.response_header["CONTENT_TYPE"]
    end

    def headers
      @http.response_header
    end

    def stream(&block)
      if block
        block.call(@body)

        @http.stream(&block)
        EM::Synchrony::Iterator.new([block]).each do |_, iter|
          @http.callback { iter.next }
        end
      else
        fail "a block is required when streaming"
      end
    end
  end
end
