module ApiTransformer
  # Processes endpoint blocks
  class Endpoint
    def initialize(base_url, env, route)
      @base_url = base_url
      @env = env
      @route = route

      @backend_request_senders = []
      @backend_responses = {}
    end

    def run
      fail "Endpoints must define a response block" unless @builder

      send_requests
      send_response
    end

    def complete
      if @error
        complete_error
      else
        complete_success
      end
    end

    def request(name, options = {}, &block)
      return unless !options.key?(:when) || options[:when]

      @backend_request_senders << BackendRequestSender.new(
        name,
        options.merge(base_url: @base_url),
        block,
        @env["client-headers"],
        @route[:helper_blocks]
      )
    end

    def response(options = {}, &block)
      @builder = FrontendResponseBuilder.new(
        @env,
        options,
        block,
        @route
      )
    end

    def stream(value)
      @stream = value
    end

    private

    def send_requests
      @multi = EventMachine::MultiRequest.new
      iterator = EM::Synchrony::Iterator.new(@backend_request_senders)

      iterator.each do |sender, iter|
        if @backend_responses.values.all?(&:success?)
          send_request(sender, iter)
        else
          iter.next
        end
      end
    end

    def send_request(sender, iter)
      http = sender.send(@backend_responses)
      @backend_responses[sender.request_name] = BackendResponse.new(http)
      @multi.add sender.request_name, http
      http.headers { iter.next }
    rescue ApiTransformer::RequestError => e
      @error = e
      iter.next
    end

    def send_response
      if @stream
        status_and_headers
      else
        if @backend_request_senders.any? && !@error
          EM::Synchrony::Iterator.new([@multi]).each do |multi, iter|
            multi.callback { iter.next }
          end
        end

        status_and_headers
      end
    end

    def status_and_headers
      if @error
        [400, {}]
      else
        @builder.status_and_headers(@backend_responses)
      end
    end

    def complete_success
      EM.synchrony do
        @builder.do_streaming

        if @backend_request_senders.any?
          @multi.callback do
            close
          end
        else
          close
        end
      end
    end

    def complete_error
      @builder.stream_write @error
      close
    end

    def close
      @builder.send_body
      @builder.stream_close
    end
  end
end
