module ApiTransformer
  # Processes the response block
  class FrontendResponseBuilder
    attr_reader :success

    def initialize(env, options, block, route)
      @env = env
      @options = options
      @block = block
      @route = route

      @frontend_response = FrontendResponse.new
    end

    def status_and_headers(backend_responses)
      @route[:helper_blocks].each { |block| instance_eval(&block) }
      instance_exec(backend_responses, &@block)
      add_failure_handlers(backend_responses)
      handle_success_or_failure(backend_responses)

      [@frontend_response.status, @frontend_response.headers]
    end

    def body(value)
      @frontend_response.body = value
    end

    def streaming?
      options[:streaming]
    end

    def success(value = nil, &block)
      @frontend_response.status = value
      @success = block
    end

    def failure(run = true, &block)
      run && !@failure && @failure = block
    end

    def status(value)
      @frontend_response.status = value
    end

    def content_type(value)
      @frontend_response.content_type = value
    end

    def cookie(key, value)
      @frontend_response.set_cookie(key, value)
    end

    def header(key, value)
      @frontend_response.set_header(key, value)
    end

    def attribute(key, value)
      @frontend_response.set key, value
    end

    def object(key, klass, object_data)
      @frontend_response.set key, object_hash(klass, object_data)
    end

    def array(key, klass, array_data)
      value = array_data.map { |data| object_hash(klass, data) }
      @frontend_response.set key, value
    end

    def send_body
      stream_write @frontend_response.body
    end

    def stream(&block)
      if block
        @streaming_block = block
      else
        fail "a block is required when streaming"
      end
    end

    def do_streaming
      @streaming_block.call if @streaming_block
    end

    def stream_write(data)
      @env.stream_send data
    end

    def stream_close
      @env.stream_close
    end

    private

    def add_failure_handlers(backend_responses)
      @route[:failure_handlers].each do |block|
        backend_responses.values.each do |backend_response|
          instance_exec(backend_response, &block)
        end
      end
    end

    def handle_success_or_failure(backend_responses)
      if @failure
        handle_failure
      elsif backend_responses.all? { |_, resp| resp.success? }
        handle_success
      else
        unhandled_failure(backend_responses)
      end
    end

    def handle_failure
      set_failure_defaults
      instance_eval(&@failure) if @failure
    end

    def handle_success
      set_success_defaults
      instance_eval(&@success) if @success
    end

    def unhandled_failure(backend_responses)
      first_failure = backend_responses.values.first { |resp| !resp.success }
      status first_failure.status
      @frontend_response.body = first_failure.body
    end

    def object_hash(klass, data)
      klass.new(data).to_hash
    end

    def set_success_defaults
      @frontend_response.status ||= 200
    end

    def set_failure_defaults
      @frontend_response.status ||= 400
    end
  end
end
